package cn.javacoder.test_spring_aop;

import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Pointcut;
import org.aspectj.lang.reflect.MethodSignature;


@Aspect
public class TimeElapseRecordingAdvice implements MethodInterceptor {
	private static ThreadLocal<Map<String, Object>> threadLocal = null;
	
	private static class Pair {
		public int count =0;
		public long timeElpasedInMillSec = 0;
	}
	
	@Pointcut("@annotation(cn.javacoder.test_spring_aop.TimeElapse)")
	public void pointcut(){
	}
	
	@Around("pointcut()")
	public void recordTimeElapsed(ProceedingJoinPoint joinpoint) throws IOException {
		String signature = "TimeElapseRecordingAspect";
		boolean isAnnotatedWithTimeElapse = false;
		try {
			if (joinpoint.getSignature() instanceof MethodSignature) {
				MethodSignature methodSignature = (MethodSignature) joinpoint
						.getSignature();
				TimeElapse timeElaspe = methodSignature
						.getMethod()
						.getAnnotation(TimeElapse.class);
				if (timeElaspe != null && timeElaspe.needInit()) {
					threadLocal = new ThreadLocal<Map<String, Object>>();
					threadLocal.set(new LinkedHashMap<String, Object>());
					isAnnotatedWithTimeElapse = true;
				}
			}
			String typeName = joinpoint.getSignature().getDeclaringTypeName();
			String methodName = joinpoint.getSignature().getName();
			signature = typeName + "." + methodName;
		} catch (Exception e) {
			e.printStackTrace();
		}
		try {
			long start = System.currentTimeMillis();
			joinpoint.proceed();
			long end = System.currentTimeMillis();
			if(threadLocal != null) {
				Pair pair = (Pair)threadLocal.get().get(signature);
				if(pair != null) {
					pair.count += 1;
					pair.timeElpasedInMillSec += (end - start);
				} else {
					pair = new Pair();
					pair.count = 1;
					pair.timeElpasedInMillSec = (end - start);
					threadLocal.get().put(signature, pair);
				}
			}
		} catch (Throwable t) {
			t.printStackTrace();
		} finally {
			if(isAnnotatedWithTimeElapse ) {
				StringWriter writer = new StringWriter();
				if(threadLocal != null) {
					List<String> temp = new ArrayList<String>();
					for(Map.Entry<String, Object> entry : threadLocal.get().entrySet()) {
						Pair pair = (Pair)entry.getValue();
						temp.add(entry.getKey() + "\tcount:" + pair.count + "\ttime elapsed:" + pair.timeElpasedInMillSec + "\r\n");
					}
					for(int i = temp.size()-1; i>=0; i--) {
						writer.write(temp.get(i));
					}
				}
				System.out.println(writer.toString());
				isAnnotatedWithTimeElapse = false;
				threadLocal = null;
			}
		}
	}
	

	public Object invoke(MethodInvocation invocation) throws Throwable {
		Object retValue = null;
		String signature = "TimeElapseRecordingAspect";
		boolean isAnnotatedWithTimeElapse = false;
		try {
			TimeElapse timeElaspe = invocation.getMethod().getAnnotation(TimeElapse.class);
			if (timeElaspe != null && timeElaspe.needInit()) {
				threadLocal = new ThreadLocal<Map<String, Object>>();
				threadLocal.set(new LinkedHashMap<String, Object>());
				isAnnotatedWithTimeElapse = true;
			}

			String typeName = invocation.getMethod().getDeclaringClass().getCanonicalName();
			String methodName = invocation.getMethod().getName();
			signature = typeName + "." + methodName;
		} catch (Exception e) {
			e.printStackTrace();
		}
		try {
			long start = System.currentTimeMillis();
			retValue = invocation.proceed();
			long end = System.currentTimeMillis();
			if (threadLocal != null && threadLocal.get() != null
					&& !signature.contains("alibaba.dubbo.common.bytecode")) {
				Pair pair = (Pair) threadLocal.get().get(signature);
				if (pair != null) {
					pair.count += 1;
					pair.timeElpasedInMillSec += (end - start);
				} else {
					pair = new Pair();
					pair.count = 1;
					pair.timeElpasedInMillSec = (end - start);
					threadLocal.get().put(signature, pair);
				}
			}
		} catch (Throwable t) {
			t.printStackTrace();
		} finally {
			if (isAnnotatedWithTimeElapse) {
				try {

					StringWriter writer = new StringWriter();
					if (threadLocal != null) {
						List<String> temp = new ArrayList<String>();
						for (Map.Entry<String, Object> entry : threadLocal.get().entrySet()) {
							Pair pair = (Pair) entry.getValue();
							temp.add("\n" + entry.getKey() + "\tcount:" + pair.count + "\ttime elapsed:"
									+ pair.timeElpasedInMillSec);
						}
						for (int i = temp.size() - 1; i >= 0; i--) {
							writer.write(temp.get(i));
						}
						writer.write("\n");
					}
					System.out.println(writer.toString());
					isAnnotatedWithTimeElapse = false;
					threadLocal.set(null);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
		return retValue;
	}
}
